@@ -1,4 +1,9 @@
-"""Auxiliary services for working with Python type annotations."""
+"""
+Type-safe data interchange for Python data classes.
+
+Provides auxiliary services for working with Python type annotations, converting typed data to and from JSON,
+and generating a JSON schema for a complex type.
+"""
 
 import base64
 import dataclasses
@@ -8,7 +13,9 @@ import inspect
 import json
 import keyword
 import re
+import types
 import typing
+import uuid
 from typing import (
     Any,
     ClassVar,
@@ -33,6 +40,8 @@ OBJECT_ENUM_EXPANSION_LIMIT = 4
 JsonType = Union[None, bool, int, float, str, Dict[str, "JsonType"], List["JsonType"]]
 Schema = JsonType
 T = TypeVar("T")
+K = TypeVar("K")
+V = TypeVar("V")
 
 
 def is_dataclass_type(typ) -> bool:
@@ -72,6 +81,13 @@ def is_named_tuple_type(typ) -> bool:
     return all(type(n) == str for n in f)
 
 
+def get_module_classes(module: types.ModuleType) -> List[Type]:
+    is_class_member = (
+        lambda member: inspect.isclass(member) and member.__module__ == module.__name__
+    )
+    return [class_type for _, class_type in inspect.getmembers(module, is_class_member)]
+
+
 def get_class_properties(typ: Type) -> Iterable[Tuple[str, Type]]:
     resolved_hints = typing.get_type_hints(typ)
 
@@ -114,6 +130,8 @@ def object_to_json(obj: Any) -> JsonType:
         return base64.b64encode(obj).decode("ascii")
     elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
         return obj.isoformat()
+    elif isinstance(obj, uuid.UUID):
+        return str(obj)
     elif isinstance(obj, enum.Enum):
         return obj.value
     elif isinstance(obj, list):
@@ -205,6 +223,8 @@ def json_to_object(typ: Type[T], data: JsonType) -> T:
         return base64.b64decode(data)
     elif typ is datetime.datetime or typ is datetime.date or typ is datetime.time:
         return typ.fromisoformat(data)
+    elif typ is uuid.UUID:
+        return uuid.UUID(data)
 
     # generic types (e.g. list, dict, set, etc.)
     origin_type = typing.get_origin(typ)
@@ -240,16 +260,16 @@ def json_to_object(typ: Type[T], data: JsonType) -> T:
 
         raise KeyError(f"type `{typ}` could not be instantiated from: {data}")
 
+    if not inspect.isclass(typ):
+        raise TypeError(f"unable to de-serialize unrecognized type `{typ}`")
+
     if is_named_tuple_type(typ):
         object_dict = {
             field_name: json_to_object(field_type, data[field_name])
-            for field_name, field_type in typ.__annotations__.items()
+            for field_name, field_type in typing.get_type_hints(typ).items()
         }
         return typ(**object_dict)
 
-    if not inspect.isclass(typ):
-        raise TypeError(f"unable to de-serialize unrecognized type `{typ}`")
-
     if issubclass(typ, enum.Enum):
         return typ(data)
 
@@ -287,6 +307,13 @@ def json_dump_string(json_object: JsonType) -> str:
     )
 
 
+def is_type_enum(typ: Type) -> bool:
+    "True if the specified type is an enumeration type."
+
+    # use an explicit isinstance(..., type) check to filter out special forms like generics
+    return isinstance(typ, type) and issubclass(typ, enum.Enum)
+
+
 def is_type_optional(typ: Type) -> bool:
     "True if the type annotation corresponds to an optional type (e.g. Optional[T] or Union[T1,T2,None])."
 
@@ -305,46 +332,100 @@ def unwrap_optional_type(typ: Type[Optional[T]]) -> Type[T]:
 
     # will automatically unwrap Union[T] into T
     return Union[
-        tuple(filter(lambda item: item is not type(None), typing.get_args(typ)))
+        tuple(filter(lambda item: item is not type(None), typing.get_args(typ)))  # type: ignore
     ]
 
 
+def is_generic_list(typ: Type) -> bool:
+    "True if the specified type is a generic list, i.e. `List[T]`."
+
+    return typing.get_origin(typ) is list
+
+
+def unwrap_generic_list(typ: Type[List[T]]) -> Type[T]:
+    (list_type,) = typing.get_args(typ)  # unpack single tuple element
+    return list_type
+
+
+def is_generic_dict(typ: Type) -> bool:
+    "True if the specified type is a generic dictionary, i.e. `Dict[KeyType, ValueType]`."
+
+    return typing.get_origin(typ) is dict
+
+
+def unwrap_generic_dict(typ: Type[Dict[K, V]]) -> Tuple[Type[K], Type[V]]:
+    key_type, value_type = typing.get_args(typ)
+    return key_type, value_type
+
+
 try:
     import docstring_parser
 
-    def docstring_to_schema(typ: Type) -> Schema:
-        docstring = docstring_parser.parse(typ.__doc__)
-        schema = dict()
-        if docstring.short_description:
-            schema["title"] = docstring.short_description
-        if docstring.long_description:
-            schema["description"] = docstring.long_description
-        return schema
+    def get_class_docstrings(typ: Type) -> Tuple[Optional[str], Optional[str]]:
+        docstring: docstring_parser.Docstring = docstring_parser.parse(typ.__doc__)
+        return docstring.short_description, docstring.long_description
+
+    def get_class_property_docstrings(typ: Type) -> Dict[str, str]:
+        docstring: docstring_parser.Docstring = docstring_parser.parse(typ.__doc__)
+        return {param.arg_name: param.description for param in docstring.params}
 
 
 except ImportError:
 
-    def docstring_to_schema(typ: Type) -> Schema:
-        return dict()
+    def get_class_property_docstrings(typ: Type) -> Dict[str, str]:
+        return {}
+
+    def get_class_docstrings(typ: Type) -> Tuple[Optional[str], Optional[str]]:
+        return None, None
+
+
+def docstring_to_schema(typ: Type) -> Schema:
+    short_description, long_description = get_class_docstrings(typ)
+    schema = dict()
+    if short_description:
+        schema["title"] = short_description
+    if long_description:
+        schema["description"] = long_description
+    return schema
+
+
+@dataclasses.dataclass
+class SchemaOptions:
+    definitions_path: str = "#/definitions/"
+    use_descriptions: bool = True
 
 
 class JsonSchemaGenerator:
     """Creates a JSON schema with user-defined type definitions."""
 
     type_catalog: ClassVar[Dict[Type, Schema]] = {}
-    types_used: Set[Type]
-    use_descriptions: bool = False
+    types_used: Dict[str, Type]
+    options: SchemaOptions
 
-    def __init__(self, definitions_path: str = None, use_descriptions: bool = True):
-        if not definitions_path:
-            definitions_path = "#/definitions/"
-        self.definitions_path = definitions_path
-        self.use_descriptions = use_descriptions
-        self.types_used = set()
+    def __init__(self, options: SchemaOptions = None):
+        if options is None:
+            self.options = SchemaOptions
+        else:
+            self.options = options
+        self.types_used = {}
 
     def type_to_schema(self, typ: Type, force_expand: bool = False) -> Schema:
+        if isinstance(typ, str):
+            raise TypeError(f"object is not a type but a string")
+
         if typ is None:
             return {"type": "null"}
+        elif typ is Any:
+            return {
+                "anyOf": [
+                    {"type": "null"},
+                    {"type": "boolean"},
+                    {"type": "number"},
+                    {"type": "string"},
+                    {"type": "array"},
+                    {"type": "object"},
+                ]
+            }
         elif typ is bool:
             return {"type": "boolean"}
         elif typ is int:
@@ -367,7 +448,18 @@ class JsonSchemaGenerator:
         elif typ is datetime.time:
             # 20:20:39+00:00
             return {"type": "string", "format": "time"}
-        elif inspect.isclass(typ) and issubclass(typ, enum.Enum):
+        elif typ is uuid.UUID:
+            # f81d4fae-7dec-11d0-a765-00a0c91e6bf6
+            return {"type": "string", "format": "uuid"}
+
+        if isinstance(typ, typing.ForwardRef):
+            fwd: typing.ForwardRef = typ
+            identifier = fwd.__forward_arg__
+            typ = eval(identifier)
+            self.types_used[identifier] = typ
+            return {"$ref": f"{self.options.definitions_path}{identifier}"}
+
+        if is_type_enum(typ):
             enum_values = [e.value for e in typ]
             enum_value_types = list(dict.fromkeys(type(value) for value in enum_values))
             if len(enum_value_types) != 1:
@@ -390,7 +482,7 @@ class JsonSchemaGenerator:
                 )
 
             enum_schema = {"type": enum_schema_type, "enum": enum_values}
-            if self.use_descriptions:
+            if self.options.use_descriptions:
                 enum_schema.update(docstring_to_schema(typ))
             return enum_schema
 
@@ -400,15 +492,13 @@ class JsonSchemaGenerator:
             return {"type": "array", "items": self.type_to_schema(list_type)}
         elif origin_type is dict:
             key_type, value_type = typing.get_args(typ)
-            if not (
-                key_type is str or key_type is int or issubclass(key_type, enum.Enum)
-            ):
+            if not (key_type is str or key_type is int or is_type_enum(key_type)):
                 raise ValueError(
                     "`dict` with key type not coercible to `str` is not supported"
                 )
 
             value_schema = self.type_to_schema(value_type)
-            if issubclass(key_type, enum.Enum):
+            if is_type_enum(key_type):
                 enum_values = [e.value for e in key_type]
                 if len(enum_values) > OBJECT_ENUM_EXPANSION_LIMIT:
                     dict_schema = {
@@ -447,6 +537,9 @@ class JsonSchemaGenerator:
                     self.type_to_schema(member_type) for member_type in args
                 ],
             }
+        elif origin_type is type:
+            (concrete_type,) = typing.get_args(typ)  # unpack single tuple element
+            return {"const": self.type_to_schema(concrete_type, force_expand=True)}
         elif origin_type is Union:
             return {
                 "oneOf": [
@@ -455,14 +548,17 @@ class JsonSchemaGenerator:
                 ]
             }
 
-        if not force_expand and typ in JsonSchemaGenerator.type_catalog:
+        if not force_expand and typ in __class__.type_catalog:
             # user-defined type
-            self.types_used.add(typ)
-            return {"$ref": f"{self.definitions_path}{typ.__name__}"}
+            identifier = typ.__name__
+            self.types_used[identifier] = typ
+            return {"$ref": f"{self.options.definitions_path}{identifier}"}
         else:
             # dictionary of class attributes
             members = dict(inspect.getmembers(typ, lambda a: not inspect.isroutine(a)))
 
+            property_docstrings = get_class_property_docstrings(typ)
+
             properties: Dict[str, Schema] = {}
             required: List[Schema] = []
             for property_name, property_type in get_class_properties(typ):
@@ -492,6 +588,12 @@ class JsonSchemaGenerator:
                     ):
                         property_def["default"] = object_to_json(def_value)
 
+                # add property docstring if available
+                property_doc = property_docstrings.get(property_name)
+                if property_doc:
+                    property_def["title"] = property_doc
+                    property_def.pop("description", None)
+
                 properties[property_name] = property_def
 
             schema = {"type": "object"}
@@ -500,17 +602,17 @@ class JsonSchemaGenerator:
                 schema["additionalProperties"] = False
             if len(required) > 0:
                 schema["required"] = required
-            if self.use_descriptions:
+            if self.options.use_descriptions:
                 schema.update(docstring_to_schema(typ))
             return schema
 
     def _subtype_to_schema(self, subtype: Type) -> Schema:
-        subschema = JsonSchemaGenerator.type_catalog[subtype]
+        subschema = __class__.type_catalog[subtype]
         if subschema is not None:
             type_schema = subschema.copy()
 
             # add descriptive text (if present)
-            if self.use_descriptions:
+            if self.options.use_descriptions:
                 type_schema.update(docstring_to_schema(subtype))
         else:
             type_schema = self.type_to_schema(subtype, force_expand=True)
@@ -518,38 +620,38 @@ class JsonSchemaGenerator:
         return type_schema
 
     def classdef_to_schema(self, typ: Type) -> Tuple[Schema, Dict[str, Schema]]:
-        self.types_used = set()
+        self.types_used = {}
         try:
             type_schema = self.type_to_schema(typ)
 
-            types_defined = set()
-            subtype_mapping = {}
+            types_defined = {}
             while len(self.types_used) > len(types_defined):
-                types_undefined = self.types_used - types_defined
-                for subtype in types_undefined:
-                    subtype_mapping[subtype] = self._subtype_to_schema(subtype)
-                types_defined.update(types_undefined)
+                # make a snapshot copy; original collection is going to be modified
+                types_undefined = {
+                    sub_name: sub_type
+                    for sub_name, sub_type in self.types_used.items()
+                    if sub_name not in types_defined
+                }
 
-            type_definitions = dict(
-                sorted(
-                    (subtype.__name__, subschema)
-                    for subtype, subschema in subtype_mapping.items()
-                )
-            )
+                # expand undefined types, which may lead to additional types to be defined
+                for sub_name, sub_type in types_undefined.items():
+                    types_defined[sub_name] = self._subtype_to_schema(sub_type)
+
+            type_definitions = dict(sorted(types_defined.items()))
         finally:
-            self.types_used = set()
+            self.types_used = {}
 
         return type_schema, type_definitions
 
 
-def classdef_to_schema(typ: Type) -> Schema:
+def classdef_to_schema(typ: Type, options: SchemaOptions = None) -> Schema:
     """Returns the JSON schema corresponding to the given type.
 
     :param typ: The Python type used to generate the JSON schema
     :return: A JSON object that you can serialize to a JSON string with json.dump or json.dumps
     """
 
-    type_schema, type_definitions = JsonSchemaGenerator().classdef_to_schema(typ)
+    type_schema, type_definitions = JsonSchemaGenerator(options).classdef_to_schema(typ)
 
     schema = {"$schema": "http://json-schema.org/draft-07/schema#"}
     if type_definitions:
@@ -579,7 +681,7 @@ def print_schema(typ: Type) -> None:
     print(json.dumps(s, indent=4))
 
 
-def _register_schema(cls, schema=None):
+def register_schema(cls, schema=None):
     JsonSchemaGenerator.type_catalog[cls] = schema
     return cls
 
@@ -588,7 +690,7 @@ def json_schema_type(cls=None, /, *, schema=None):
     """Decorator to add user-defined schema definition to a class."""
 
     def wrap(cls):
-        return _register_schema(cls, schema)
+        return register_schema(cls, schema)
 
     # see if decorator is used as @json_schema_type or @json_schema_type()
     if cls is None:
@@ -597,3 +699,6 @@ def json_schema_type(cls=None, /, *, schema=None):
     else:
         # called as @json_schema_type without parentheses
         return wrap(cls)
+
+
+register_schema(JsonType)
